use std::marker::PhantomData;

use avian2d::prelude::*;
use bevy::prelude::*;
pub use here_be_dragons::Map as MapgenMap;
use here_be_dragons::{geometry::Rect as MRect, MapFilter, Tile};
use maze_generator::{prelude::*, recursive_backtracking::RbGenerator};
use pathfinding::prelude::*;
use rand::prelude::StdRng;

use crate::{
    core::{Obstacle, PointLike, Zone},
    exploration::Mappable,
    pathfinding::CostMap,
    visibility::Visible,
};

#[derive(Component, Clone, Default, Deref, DerefMut)]
#[require(Transform, SpawnColliders, SpawnPortals)]
pub struct Map<D: 'static + Clone + Default + Send + Sync>(pub MapgenMap<D>);

impl<D: 'static + Clone + Default + Send + Sync> Map<D> {
    pub fn find_path(
        &self,
        start: &UVec2,
        destination: &UVec2,
        cost_map: &Option<CostMap>,
    ) -> Option<(Vec<UVec2>, u32)> {
        astar(
            &start,
            |p| {
                let mut successors: Vec<(UVec2, u32)> = vec![];
                if let Some(tile) = self.at(p.x as usize, p.y as usize) {
                    if tile.is_walkable() {
                        for tile in self.get_available_exits(p.x as usize, p.y as usize) {
                            let mut cost = tile.2 * 100.;
                            if let Some(cost_map) = cost_map {
                                if let Some(modifier) = cost_map.get(&p.as_ivec2()) {
                                    cost *= modifier;
                                }
                            }
                            successors
                                .push((UVec2::new(tile.0 as u32, tile.1 as u32), cost as u32));
                        }
                    }
                }
                successors
            },
            |p| (p.distance_squared(destination) * 100.) as u32,
            |p| *p == *destination,
        )
    }
}

impl<D: Clone + Default + Send + Sync> From<MapgenMap<D>> for Map<D> {
    fn from(v: MapgenMap<D>) -> Self {
        Self(v)
    }
}

#[derive(Component, Clone, Debug, Default, Reflect)]
#[reflect(Component)]
#[require(Transform, Mappable)]
pub struct Portal;

#[derive(Component, Clone, Debug, Deref, DerefMut, Reflect)]
#[reflect(Component)]
pub struct SpawnColliders(pub bool);

impl Default for SpawnColliders {
    fn default() -> Self {
        Self(true)
    }
}

#[derive(Component, Clone, Debug, Deref, DerefMut, Reflect)]
#[reflect(Component)]
pub struct SpawnPortals(pub bool);

impl Default for SpawnPortals {
    fn default() -> Self {
        Self(true)
    }
}

pub trait ITileType {
    fn blocks_motion(&self) -> bool;
    fn blocks_visibility(&self) -> bool;
}

impl ITileType for Tile {
    fn blocks_motion(&self) -> bool {
        self.is_blocked()
    }

    fn blocks_visibility(&self) -> bool {
        self.is_blocked()
    }
}

pub struct GridBuilder {
    width_in_rooms: usize,
    height_in_rooms: usize,
    room_width: usize,
    room_height: usize,
}

impl GridBuilder {
    pub fn new(
        width_in_rooms: usize,
        height_in_rooms: usize,
        room_width: usize,
        room_height: usize,
    ) -> Box<GridBuilder> {
        Box::new(GridBuilder {
            width_in_rooms,
            height_in_rooms,
            room_width,
            room_height,
        })
    }
}

impl<D: Clone + Default> MapFilter<D> for GridBuilder {
    fn modify_map(&self, _rng: &mut StdRng, map: &MapgenMap<D>) -> MapgenMap<D> {
        let mut map = map.clone();
        let mut generator = RbGenerator::new(None);
        if let Ok(maze) =
            generator.generate(self.width_in_rooms as i32, self.height_in_rooms as i32)
        {
            let total_height = self.room_height * self.height_in_rooms + self.height_in_rooms + 1;
            let half_width = self.room_width / 2;
            let half_height = self.room_height / 2;
            for y in 0..self.height_in_rooms {
                for x in 0..self.width_in_rooms {
                    // println!("({x}, {y}): ");
                    let x_offset = x * (self.room_width + 1);
                    let y_offset = total_height - (y + 1) * (self.room_height + 1) - 1;
                    // println!("Offsets: {x_offset}, {y_offset}");
                    let room = MRect::new_i32(
                        x_offset as i32 + 1,
                        y_offset as i32 + 1,
                        self.room_width as i32,
                        self.room_height as i32,
                    );
                    // println!("{room:?}");
                    map.add_room(room);
                    let coords = maze_generator::prelude::Coordinates::new(x as i32, y as i32);
                    if let Some(field) = maze.get_field(&coords) {
                        use maze_generator::prelude::Direction::*;
                        if field.has_passage(&North) {
                            let x = x_offset + half_width;
                            let y = y_offset + self.room_height + 1;
                            map.set_tile(x, y, Tile::floor());
                        }
                        if field.has_passage(&South) {
                            let x = x_offset + half_width;
                            let y = y_offset;
                            map.set_tile(x, y, Tile::floor());
                        }
                        if field.has_passage(&East) {
                            let x = x_offset + self.room_width + 1;
                            let y = y_offset + half_height;
                            map.set_tile(x, y, Tile::floor());
                        }
                        if field.has_passage(&West) {
                            let x = x_offset;
                            let y = y_offset + half_height;
                            map.set_tile(x, y, Tile::floor());
                        }
                    }
                }
            }
            map.starting_point = Some(here_be_dragons::geometry::Point::new(
                maze.start.x as usize + 1,
                maze.start.y as usize + 1,
            ));
            map.exit_point = Some(here_be_dragons::geometry::Point::new(
                (maze.goal.x as usize) * self.room_width + half_width,
                (maze.goal.y as usize) * self.room_height + half_height,
            ));
        }
        map
    }
}

fn spawn_colliders<D: 'static + Clone + Default + Send + Sync>(
    mut commands: Commands,
    maps: Query<(Entity, &Map<D>, &SpawnColliders), Changed<SpawnColliders>>,
) {
    for (map_entity, map, spawn_colliders) in &maps {
        if **spawn_colliders {
            commands.entity(map_entity).remove::<SpawnColliders>();
            for y in 0..map.height {
                for x in 0..map.width {
                    if let Some(tile) = map.at(x, y) {
                        if tile.blocks_motion() {
                            let id = commands
                                .spawn((
                                    Obstacle,
                                    Transform::from_xyz(x as f32 + 0.5, y as f32 + 0.5, 0.),
                                ))
                                .id();
                            if tile.blocks_visibility() {
                                commands.entity(id).insert(Visible::opaque());
                            }
                            commands.entity(map_entity).add_children(&[id]);
                        }
                    }
                }
            }
            for room in &map.rooms {
                commands.entity(map_entity).with_children(|parent| {
                    parent.spawn((
                        Zone,
                        Transform::from_xyz(room.center().x as f32, room.center().y as f32, 0.),
                        Collider::rectangle(room.width() as f32 + 1., room.height() as f32 + 1.),
                    ));
                });
            }
        }
    }
}

fn spawn_portals<D: 'static + Clone + Default + Send + Sync>(
    mut commands: Commands,
    map: Query<(Entity, &Map<D>, &SpawnPortals), Changed<SpawnPortals>>,
) {
    for (map_entity, map, spawn_portals) in &map {
        if **spawn_portals {
            commands.entity(map_entity).remove::<SpawnPortals>();
            let mut portals: Vec<(f32, f32)> = vec![];
            for x in 0..map.width {
                for y in 0..map.height {
                    let idx = (x, y).to_index(map.width);
                    let mut spawn_portal = false;
                    if map.get_available_exits(x, y).len() > 2 && map.tiles[idx].is_walkable() {
                        if (x > 1 && map.tiles[idx - 1].is_walkable())
                            && (x < map.width - 2 && map.tiles[idx + 1].is_walkable())
                            && (y > 1 && map.tiles[idx - map.width].is_blocked())
                            && (y < map.height - 2 && map.tiles[idx + map.width].is_blocked())
                        {
                            spawn_portal = true;
                        }
                        if (x > 1 && map.tiles[idx - 1].is_blocked())
                            && (x < map.width - 2 && map.tiles[idx + 1].is_blocked())
                            && (y > 1 && map.tiles[idx - map.width].is_walkable())
                            && (y < map.height - 2 && map.tiles[idx + map.width].is_walkable())
                        {
                            spawn_portal = true;
                        }
                    }
                    if spawn_portal {
                        let x = x as f32 + 0.5;
                        let y = y as f32 + 0.5;
                        if !portals.contains(&(x, y)) {
                            portals.push((x, y));
                        }
                    }
                }
            }
            for (x, y) in portals {
                commands.entity(map_entity).with_children(|parent| {
                    parent.spawn((Portal, Transform::from_xyz(x, y, 0.)));
                });
            }
        }
    }
}

fn spawn_portal_colliders<D: 'static + Clone + Default + Send + Sync>(
    mut commands: Commands,
    map: Query<(Entity, &SpawnColliders), With<Map<D>>>,
    portals: Query<Entity, (With<Portal>, Without<Collider>)>,
) {
    for (entity, spawn_colliders) in &map {
        if **spawn_colliders {
            for portal_entity in &portals {
                commands
                    .entity(portal_entity)
                    .insert((Collider::rectangle(1., 1.), Sensor));
            }
            commands.entity(entity).remove::<SpawnColliders>();
        }
    }
}

#[derive(Resource, Clone, Copy, Debug)]
pub struct MapPlugin<MapData: 'static + Clone + Default + Send + Sync> {
    pub start_revealed: bool,
    pub map_data: PhantomData<MapData>,
}

impl<MapData: 'static + Clone + Default + Send + Sync> Default for MapPlugin<MapData> {
    fn default() -> Self {
        Self {
            start_revealed: default(),
            map_data: default(),
        }
    }
}

impl<MapData: 'static + Clone + Default + Send + Sync> Plugin for MapPlugin<MapData> {
    fn build(&self, app: &mut App) {
        app.insert_resource(self.clone())
            .register_type::<Portal>()
            .add_systems(
                PreUpdate,
                (
                    spawn_colliders::<MapData>,
                    spawn_portals::<MapData>,
                    spawn_portal_colliders::<MapData>,
                ),
            );
    }
}
